from typing import (
    TYPE_CHECKING,
    Union,
)

from galaxy.exceptions import (
    error_codes,
    MessageException,
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.util.json import safe_dumps

if TYPE_CHECKING:
    from fastapi.exceptions import RequestValidationError
    from pydantic import ValidationError


def validation_error_to_message_exception(e: Union["ValidationError", "RequestValidationError"]) -> MessageException:
    invalid_found = False
    missing_found = False
    messages = []
    clean_validation_errors = []
    for error in e.errors():
        messages.append(f"{error['msg']} in {error['loc']}")
        if error["type"] == "message_exception" and "ctx" in error:
            return error["ctx"]["exception"]
        if error["type"] == "missing" or error["type"] == "type_error.none.not_allowed":
            missing_found = True
        elif error["type"].startswith("type_error"):
            invalid_found = True
        # ctx contains data that can't be serialized, like exception instances
        error.pop("ctx", None)
        try:
            clean_validation_errors.append(safe_dumps(error))
        except TypeError:
            pass
    if missing_found and not invalid_found:
        return RequestParameterMissingException("\n".join(messages), validation_errors=clean_validation_errors)
    else:
        return RequestParameterInvalidException("\n".join(messages), validation_errors=clean_validation_errors)


def api_error_to_dict(**kwds):
    UNKNOWN_ERROR_CODE = error_codes.error_codes_by_name["UNKNOWN"]
    exception = kwds.get("exception", None)
    if exception:
        # If we are passed a MessageException use err_msg.
        default_error_code = getattr(exception, "err_code", UNKNOWN_ERROR_CODE)
        default_error_message = getattr(exception, "err_msg", default_error_code.default_error_message)
        extra_error_info = getattr(exception, "extra_error_info", {})
        if not isinstance(extra_error_info, dict):
            extra_error_info = {}
    else:
        default_error_message = "Error processing API request."
        default_error_code = UNKNOWN_ERROR_CODE
        extra_error_info = {}
    traceback_string = kwds.get("traceback", "No traceback available.")
    err_msg = kwds.get("err_msg", default_error_message)
    error_code_object = kwds.get("err_code", default_error_code)
    try:
        error_code = error_code_object.code
    except AttributeError:
        # Some sort of bad error code sent in, logic failure on part of
        # Galaxy developer.
        error_code = UNKNOWN_ERROR_CODE.code
    # Would prefer the terminology of error_code and error_message, but
    # err_msg used a good number of places already. Might as well not change
    # it?
    error_response = dict(err_msg=err_msg, err_code=error_code, **extra_error_info)
    if kwds.get("debug"):  # TODO: Should admins get to see traceback as well?
        error_response["traceback"] = traceback_string
    return error_response



import os
from datetime import datetime
from functools import wraps
from typing import (
    Callable,
    TypeVar,
    Union,
)
from unittest import SkipTest

import pytest
from typing_extensions import ParamSpec

from galaxy.util import requests
from galaxy.util.commands import which


def is_site_up(url: str) -> bool:
    try:
        response = requests.get(url, timeout=10)
        return response.status_code == 200
    except Exception:
        return False


P = ParamSpec("P")
T = TypeVar("T")


def skip_if_site_down(url: str) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def method_wrapper(method: Callable[P, T]) -> Callable[P, T]:
        @wraps(method)
        def wrapped_method(*args: P.args, **kwargs: P.kwargs) -> T:
            if not is_site_up(url):
                raise SkipTest(f"Test depends on [{url}] being up and it appears to be down.")
            return method(*args, **kwargs)

        return wrapped_method

    return method_wrapper


skip_if_github_down = skip_if_site_down("https://github.com/")
skip_if_workflowhub_down = skip_if_site_down("https://workflowhub.eu/")


def _identity(func: Callable[P, T]) -> Callable[P, T]:
    return func


def skip_unless_executable(executable: str) -> Union[Callable[[Callable[P, T]], Callable[P, T]], pytest.MarkDecorator]:
    if which(executable):
        return _identity
    return pytest.mark.skip(f"PATH doesn't contain executable {executable}")


def skip_unless_environ(env_var: str) -> Union[Callable[[Callable[P, T]], Callable[P, T]], pytest.MarkDecorator]:
    if os.environ.get(env_var):
        return _identity

    return pytest.mark.skip(f"{env_var} must be set for this test")


# Pytest mark for tests that require a live LLM connection
# Set GALAXY_TEST_ENABLE_LIVE_LLM=1 to run these tests
pytestmark_live_llm = pytest.mark.skipif(
    not os.environ.get("GALAXY_TEST_ENABLE_LIVE_LLM"),
    reason="Live LLM tests disabled. Set GALAXY_TEST_ENABLE_LIVE_LLM=1 to enable.",
)


def transient_failure(issue: int, potentially_fixed: bool = False) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Mark test as known transient failure with GitHub issue tracking.

    This decorator catches exceptions from tests and rewraps them with a marker
    indicating this is a known transient failure. This allows automated tooling
    to categorize failures and helps reviewers quickly identify flaky tests.

    Please create an issue on Github to track each transient failure.

    If a potential fix is implemented, set potentially_fixed=True to
    indicate that the failure may have been resolved. This will update the
    displayed error message and help us know if the issue can be potentially
    closed after a month of not being reported.

    Args:
        issue: GitHub issue number tracking this transient failure
        potentially_fixed: If True, indicates that the underlying issue may have been fixed,
            and the test failure comment will require the PR reviewer to report
            potential failures on the tracking issue and remove potentially_fixed.

    Example:
        @transient_failure(issue=12345)
        def test_flaky_selenium(self):
            # Test that sometimes fails due to race condition
            ...
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if potentially_fixed:
                    current_datetime = datetime.now().isoformat()
                    report_this = (
                        "We have previously implemented a potential fix for this issue, "
                        "if you are seeing this failure in CI on a recently branched commit, please report it on the tracking issue "
                        f"https://github.com/galaxyproject/galaxy/issues/{issue} including the comment "
                        f"'This issue is not fixed and was last seen at {current_datetime}' so we can mark the previous fix as insufficient."
                    )
                else:
                    report_this = "This is known issue and doesn't need to be reported."

                msg = f"KNOWN TRANSIENT FAILURE [Issue #{issue}] [{report_this}]: {str(e)}"
                # Try to preserve exception type, fallback to plain Exception
                try:
                    raise type(e)(msg) from e
                except (TypeError, AttributeError):
                    # type(e) doesn't accept single string arg
                    raise Exception(msg) from e

        return wrapper

    return decorator



from typing import (
    Callable,
    cast,
    TypeVar,
)

import requests
from requests import (  # noqa: F401
    codes as codes,
    exceptions as exceptions,
    Response as Response,
)
from typing_extensions import ParamSpec

from .user_agent import get_default_headers

Param = ParamSpec("Param")
RetType = TypeVar("RetType")


def default_user_agent_decorator(f: Callable[Param, RetType]) -> Callable[Param, RetType]:

    def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
        headers = cast(dict, kwargs.pop("headers", None) or {})
        headers.update(get_default_headers())
        is_session = f in (requests.session, requests.Session)
        if not is_session:
            kwargs["headers"] = headers
        rval = f(*args, **kwargs)
        if is_session:
            rval.headers = headers  # type: ignore[attr-defined]
        return rval

    return wrapper


delete = default_user_agent_decorator(requests.delete)
get = default_user_agent_decorator(requests.get)
head = default_user_agent_decorator(requests.head)
patch = default_user_agent_decorator(requests.patch)
post = default_user_agent_decorator(requests.post)
options = default_user_agent_decorator(requests.options)
put = default_user_agent_decorator(requests.put)
session = default_user_agent_decorator(requests.session)
Session = default_user_agent_decorator(requests.Session)



"""Utilities for defining user configuration bits from admin templates.

This is capturing code shared by file source templates and object store templates.
"""

import logging
import os
from collections.abc import Iterable
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlencode

import requests
import yaml
from boltons.iterutils import remap
from pydantic import (
    BaseModel,
    ConfigDict,
    create_model,
    RootModel,
    ValidationError,
)
from pydantic.fields import FieldInfo
from typing_extensions import (
    Annotated,
    Literal,
    NotRequired,
    Protocol,
    TypedDict,
)

try:
    from jinja2 import (
        StrictUndefined,
        UndefinedError,
    )
    from jinja2.nativetypes import NativeEnvironment
except ImportError:
    NativeEnvironment = None  # type: ignore[assignment, misc, unused-ignore]
    StrictUndefined = None  # type: ignore[assignment, misc, unused-ignore]
    UndefinedError = None  # type: ignore[assignment, misc, unused-ignore]

from galaxy.exceptions import (
    ObjectNotFound,
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.tool_util_models.parameter_validators import AnySafeValidatorModel
from galaxy.util import asbool

log = logging.getLogger(__name__)

TemplateVariableType = Literal["string", "path_component", "boolean", "integer"]
TemplateVariableValueType = Union[str, bool, int]
TemplateExpansion = str
MarkdownContent = str
RawTemplateConfig = Dict[str, Any]
UserDetailsDict = Dict[str, Any]
VariablesDict = Dict[str, TemplateVariableValueType]
SecretsDict = Dict[str, str]
EnvironmentDict = Dict[str, str]


class StrictModel(BaseModel):
    model_config = ConfigDict(extra="forbid", coerce_numbers_to_str=True)


class BaseTemplateVariable(StrictModel):
    name: str
    label: Optional[str] = None
    help: Optional[MarkdownContent] = None
    optional: Optional[bool] = None
    validators: Optional[Sequence[AnySafeValidatorModel]] = None


class TemplateVariableString(BaseTemplateVariable):
    type: Literal["string"]
    default: Optional[str] = None


class TemplateVariableInteger(BaseTemplateVariable):
    type: Literal["integer"]
    default: Optional[int] = None
    # add min/max


class TemplateVariablePathComponent(BaseTemplateVariable):
    type: Literal["path_component"]
    default: Optional[str] = None


class TemplateVariableBoolean(BaseTemplateVariable):
    type: Literal["boolean"]
    default: Optional[bool] = None


TemplateVariable = Union[
    TemplateVariableString, TemplateVariableInteger, TemplateVariablePathComponent, TemplateVariableBoolean
]


class TemplateSecret(StrictModel):
    name: str
    label: Optional[str] = None
    help: Optional[MarkdownContent] = None
    optional: Optional[bool] = None


class TemplateEnvironmentSecret(StrictModel):
    type: Literal["secret"]
    name: str
    vault_key: str
    default: Optional[str] = None


class TemplateEnvironmentVariable(StrictModel):
    type: Literal["variable"]
    name: str
    variable: str
    default: Optional[str] = None


TemplateEnvironmentEntry = Union[TemplateEnvironmentVariable, TemplateEnvironmentSecret]
TemplateEnvironment = RootModel[List[TemplateEnvironmentEntry]]


def _ensure_path_component(input: Any):
    input_as_string = str(input)
    if not acts_as_simple_path_component(input_as_string):
        raise Exception("Path manipulation detected, failing evaluation")
    return input


# NativeEnvironment preserves Python types
def _environment(template_start: str, template_end: str) -> NativeEnvironment:
    env = NativeEnvironment(
        variable_start_string=template_start,
        variable_end_string=template_end,
        undefined=StrictUndefined,
    )
    env.filters["ensure_path_component"] = _ensure_path_component
    env.filters["asbool"] = asbool
    return env


class TemplateConfiguration(Protocol):

    def model_dump(self) -> Dict[str, Any]:
        """Implements a pydantic model dump to build simple JSON dictionary."""

    @property
    def template_start(self) -> Optional[str]:
        """Set a custom variable start for Jinja variable substitution.

        https://stackoverflow.com/questions/12083319/add-custom-tokens-in-jinja2-e-g-somevar
        """

    @property
    def template_end(self) -> Optional[str]:
        """Set a custom variable end for Jinja variable substitution.

        https://stackoverflow.com/questions/12083319/add-custom-tokens-in-jinja2-e-g-somevar
        """


def populate_default_variables(variables: Optional[List[TemplateVariable]], variable_values: VariablesDict):
    if variables:
        for variable in variables:
            name = variable.name
            # Apply defaults only for explicitly optional variables
            if variable.optional and name not in variable_values and variable.default is not None:
                variable_values[name] = variable.default


def expand_raw_config(
    template_configuration: TemplateConfiguration,
    variables: VariablesDict,
    secrets: SecretsDict,
    user_details: UserDetailsDict,
    environment: EnvironmentDict,
) -> RawTemplateConfig:
    template_variables = {
        "variables": variables,
        "secrets": secrets,
        "user": user_details,
        "environment": environment,
    }

    return _expand_raw_config(template_configuration, template_variables)


def _expand_raw_config(
    template_configuration: TemplateConfiguration, template_variables: Dict[str, Any]
) -> RawTemplateConfig:
    template_start = template_configuration.template_start or "{{"
    template_end = template_configuration.template_end or "}}"

    def expand_template(_, key, value):
        if isinstance(value, str) and template_start in value and template_end in value:
            template = _environment(template_start, template_end).from_string(value)
            return key, template.render(**template_variables)
        return key, value

    template_model_as_json = template_configuration.model_dump()
    raw_config = remap(template_model_as_json, visit=expand_template)
    _clean_template_meta_parameters(raw_config)
    return raw_config


def merge_implicit_parameters(raw_config: RawTemplateConfig, implicit: Optional["ImplicitConfigurationParameters"]):
    if implicit:
        raw_config.update(implicit)
        raw_config.pop("oauth2_client_id", None)
        raw_config.pop("oauth2_client_secret", None)
        raw_config.pop("oauth2_scope", None)


def verify_vault_configured_if_uses_secrets(catalog, vault_configured: bool, exception_message: str) -> None:
    if _catalog_uses_secrets(catalog) and not vault_configured:
        raise Exception(exception_message)


def _catalog_uses_secrets(catalog) -> bool:
    templates = catalog.root
    for template in templates:
        if template.secrets and len(template.secrets) > 0:
            return True
    return False


def _clean_template_meta_parameters(config: RawTemplateConfig) -> RawTemplateConfig:
    # slight templating differences between what is allowed in the template definition
    # and what is allowed in the actual configuration objects we send to respective modules
    # to instantiate plugins. In particular, descriptions of how templating is done should
    # eliminated after templates have been expanded.
    meta_parameters = ["template_start", "template_end"]
    for meta_parameter in meta_parameters:
        if meta_parameter in config:
            del config[meta_parameter]
    return config


# cwl-like - convert simple dictionary to list of dictionaries for quickly
# configuring variables and secrets
def apply_syntactic_sugar(raw_templates: List[RawTemplateConfig]) -> List[RawTemplateConfig]:
    templates = []
    expanded_raw_templates = _expand_includes(raw_templates)
    for template in expanded_raw_templates:
        _force_key_to_list(template, "variables")
        _force_key_to_list(template, "secrets")
        _force_key_to_list(template, "environment")
        templates.append(template)
    return templates


def _expand_includes(raw_templates: List[RawTemplateConfig]) -> List[RawTemplateConfig]:
    expanded_raw_templates = []
    for raw_template in raw_templates:
        expanded_raw_templates.extend(_expand_include(raw_template))
    return expanded_raw_templates


def _expand_include(raw_template: RawTemplateConfig) -> List[RawTemplateConfig]:
    has_one_key = len(raw_template.keys()) == 1
    has_include = "include" in raw_template

    if has_one_key and has_include:
        include = raw_template["include"]
        with open(include) as f:
            included = yaml.safe_load(f)
            raw_templates: List[RawTemplateConfig]
            if isinstance(included, list):
                raw_templates = included
            else:
                raw_templates = [included]
            return _expand_includes(raw_templates)
    else:
        return [raw_template]


def _force_key_to_list(template: RawTemplateConfig, key: str) -> None:
    value = template.get(key, None)
    if isinstance(value, dict):
        value_as_list = []
        for key_name, key_value in value.items():
            key_value["name"] = key_name
            value_as_list.append(key_value)
        template[key] = value_as_list


class TemplateReference(Protocol):
    template_id: str
    template_version: int


class InstanceDefinition(TemplateReference, Protocol):
    variables: Dict[str, Any]
    secrets: SecretsDict


class Template(Protocol):
    @property
    def id(self) -> str: ...

    @property
    def version(self) -> int: ...

    @property
    def type(self) -> str: ...

    @property
    def variables(self) -> Optional[List[TemplateVariable]]: ...

    @property
    def secrets(self) -> Optional[List[TemplateSecret]]: ...

    @property
    def environment(self) -> Optional[List[TemplateEnvironmentEntry]]: ...


T = TypeVar("T", bound=Template, covariant=True)


def find_template(templates: List[T], instance_reference: TemplateReference, what: str) -> T:
    template_id = instance_reference.template_id
    template_version = instance_reference.template_version
    return find_template_by(templates, template_id, template_version, what)


def find_template_by(templates: List[T], template_id: str, template_version: int, what: str) -> T:
    for template in templates:
        if template.id == template_id and template.version == template_version:
            return template

    raise ObjectNotFound(f"Could not find a {what} template with id {template_id} and version {template_version}")


def _run_variable_validator(validator: AnySafeValidatorModel, value: Any, variable_name: str) -> None:
    """Run a single validator on a variable value.

    Raises RequestParameterInvalidException if validation fails.
    """
    try:
        validator.statically_validate(value)
    except ValueError as e:
        raise RequestParameterInvalidException(f"Variable '{variable_name}' failed validation: {str(e)}")


def validate_variable_types(instance: InstanceDefinition, template: Template) -> None:
    pass


def validate_defines_all_required_secrets(instance: InstanceDefinition, template: Template):
    secrets = instance.secrets
    for template_secret in template.secrets or []:
        name = template_secret.name
        is_optional = bool(template_secret.optional)
        if name not in secrets and not is_optional:
            raise RequestParameterMissingException(f"Must define secret '{name}'")


def validate_defines_all_required_variables(instance: InstanceDefinition, template: Template):
    variables = instance.variables
    for template_variable in template.variables or []:
        name = template_variable.name
        is_optional = bool(template_variable.optional)
        if name not in variables and not is_optional:
            raise RequestParameterMissingException(f"Must define variable '{name}'")


def validate_specified_datatypes(instance: InstanceDefinition, template: Template):
    secrets = instance.secrets
    for name, value in secrets.items():
        if not isinstance(value, str):
            raise RequestParameterInvalidException(f"Secret value for secret '{name}' must be of type string")
    variables = instance.variables
    validate_specified_datatypes_variables(variables, template)


def validate_specified_datatypes_variables(variables: Dict[str, Any], template: Template):
    for template_variable in template.variables or []:
        name = template_variable.name
        # Only fall back to default for optional variables
        if name in variables:
            variable_value = variables[name]
        elif template_variable.optional:
            variable_value = template_variable.default
        else:
            variable_value = None

        # Skip validation only if variable is optional, not provided, and has no default applied
        if name not in variables and template_variable.optional and template_variable.default is None:
            continue

        template_type = template_variable.type
        if template_type in ["string", "path_component"]:
            if not isinstance(variable_value, str):
                raise RequestParameterInvalidException(f"Variable value for variable '{name}' must be of type str")
            if template_type == "path_component":
                if ".." in variable_value or "/" in variable_value:
                    raise RequestParameterInvalidException(
                        f"Variable value for variable '{name}' must be simple path component, invalid characters found"
                    )
                if not acts_as_simple_path_component(variable_value):
                    raise RequestParameterInvalidException(
                        f"Variable value for variable '{name}' must be simple path component, invalid characters found"
                    )
        if template_type == "integer":
            if not _is_of_exact_type(variable_value, int):
                raise RequestParameterInvalidException(f"Variable value for variable '{name}' must be of type int")
        if template_type == "boolean":
            if not _is_of_exact_type(variable_value, bool):
                raise RequestParameterInvalidException(f"Variable value for variable '{name}' must be of type bool")

        # Run custom validators if present.
        if template_variable.validators:
            for validator in template_variable.validators:
                _run_variable_validator(validator, variable_value, name)


def validate_no_extra_secrets_defined(secrets: Dict[str, str], template: Template) -> None:
    template_secrets = secrets_as_dict(template.secrets)
    for secret in secrets.keys():
        if secret not in template_secrets:
            raise RequestParameterInvalidException(f"No secret named {secret} for this template")


def validate_no_extra_variables_defined(variables: Dict[str, Any], template: Template):
    template_variables = _variables_as_dict(template.variables)
    for variable in variables.keys():
        if variable not in template_variables:
            raise RequestParameterInvalidException(f"No variable named {variable} for this template")


def validate_secrets_and_variables(instance: InstanceDefinition, template: Template) -> None:
    validate_defines_all_required_secrets(instance, template)
    validate_defines_all_required_variables(instance, template)
    validate_specified_datatypes(instance, template)
    validate_no_extra_secrets_defined(instance.secrets, template)
    validate_no_extra_variables_defined(instance.variables, template)


def secrets_as_dict(secrets: Optional[List[TemplateSecret]]) -> Dict[str, TemplateSecret]:
    as_dict = {}
    for secret in secrets or []:
        as_dict[secret.name] = secret
    return as_dict


def _variables_as_dict(variables: Optional[List[TemplateVariable]]) -> Dict[str, TemplateVariable]:
    as_dict = {}
    for variable in variables or []:
        as_dict[variable.name] = variable
    return as_dict


def _is_of_exact_type(object: Any, target_type: Type):
    # isinstance(False, int) and False == 0 are both True in Python...
    # We are creating a DSL here that is intentionally more strict than Python
    # so we are using type() instead of isinstance and we have the test coverage
    # to ensure this is the desired behavior and remains. Think JSON typing, not
    # pythonic typing. Galaxy's internals as a Python project should not be
    # exposed here.
    return type(object) == target_type  # noqa: E721


def acts_as_simple_path_component(value: str):
    cwd = os.getcwd()
    abs_path = os.path.abspath(f"{cwd}/{value}")
    unaffected_by_normpath = os.path.normpath(abs_path) == abs_path
    if not unaffected_by_normpath:
        return False
    should_be_cwd, should_be_value = os.path.split(abs_path)
    if should_be_cwd != cwd:
        return False
    if should_be_value != value:
        return False
    return True


class PluginAspectStatus(StrictModel):
    state: Literal["ok", "not_ok", "unknown"]
    message: str

    @property
    def is_not_ok(self):
        return self.state == "not_ok"


class PluginStatus(StrictModel):
    template_definition: PluginAspectStatus
    template_settings: Optional[PluginAspectStatus] = None
    connection: Optional[PluginAspectStatus] = None
    # I would love to disambiguate connection vs auth errors but would
    # attempting to do that cause confusion. Maybe not if the user interface
    # skipped presenting the one that couldn't be disambiguated for that
    # particular plugin?

    # TODO: Fill in writable checks.
    # writable: Optional[PluginAspectStatus] = None
    oauth2_access_token_generation: Optional[PluginAspectStatus] = None


def status_template_definition(template: Optional[Template]) -> PluginAspectStatus:
    # if we found a template in the catalog, it was validated at load time. Reflect
    # this as a PluginAspectStatus
    if template:
        return PluginAspectStatus(state="ok", message="Template definition found and validates against schema")
    else:
        return PluginAspectStatus(state="not_ok", message="Template not found or not loaded")


def settings_exception_to_status(exception: Optional[Exception]) -> PluginAspectStatus:
    if exception is None:
        status = PluginAspectStatus(state="ok", message="Valid configuration resulted from supplied settings")
    elif isinstance(exception, UndefinedError):
        message = f"Problem with template definition causing invalid settings resolution, please contact admin to correct template: {exception}"
        status = PluginAspectStatus(state="not_ok", message=message)
    elif isinstance(exception, ValidationError):
        message = f"Problem with template definition causing invalid configuration, template expanded without error but resulting configuration is invalid. please contact admin to correct template: {exception}"
        status = PluginAspectStatus(state="not_ok", message=message)
    else:
        message = f"Unknown problem with resolving configuration from supplied settings: {exception}"
        status = PluginAspectStatus(state="not_ok", message=message)
    return status


def connection_exception_to_status(what: str, exception: Optional[Exception]) -> PluginAspectStatus:
    if exception is None:
        connection_status = PluginAspectStatus(state="ok", message="Valid connection resulted from supplied settings")
    else:
        message = f"Failed to connect to a {what} with supplied settings: {exception}"
        connection_status = PluginAspectStatus(state="not_ok", message=message)
    return connection_status


class OAuth2Info(StrictModel):
    authorize_url: str


class OAuth2Configuration(StrictModel):
    authorize_url: str
    token_url: str
    authorize_params: Optional[Dict[str, str]]
    scope: Optional[str] = None


ConfiguredOAuth2Sources = Dict[str, OAuth2Configuration]


class OAuth2ClientPair(StrictModel):
    client_id: str
    client_secret: str


def get_authorize_url(
    client_id_or_pair: Union[str, OAuth2ClientPair],
    config: OAuth2Configuration,
    redirect_uri: Optional[str],
    state: Optional[str] = None,
    scope: Optional[str] = None,
) -> str:
    client_id = client_id_or_pair if isinstance(client_id_or_pair, str) else client_id_or_pair.client_id
    query_data = dict(
        client_id=client_id,
        response_type="code",
    )
    if redirect_uri is not None:
        query_data["redirect_uri"] = redirect_uri
    if state is not None:
        query_data["state"] = state
    if scope is not None:
        query_data["scope"] = scope
    elif config.scope is not None:
        query_data["scope"] = config.scope
    query_data.update(config.authorize_params or {})
    query = urlencode(query_data)
    return f"{config.authorize_url}?{query}"


def get_token_from_code_raw(
    code: str, client_pair: OAuth2ClientPair, config: OAuth2Configuration, redirect_uri: Optional[str]
) -> requests.Response:
    data = {
        "code": code,
        "grant_type": "authorization_code",
        "client_id": client_pair.client_id,
        "client_secret": client_pair.client_secret,
    }
    if redirect_uri is not None:
        data["redirect_uri"] = redirect_uri

    return requests.post(config.token_url, data=data)


def get_token_from_refresh_raw(
    refresh_token: str, client_pair: OAuth2ClientPair, config: OAuth2Configuration
) -> requests.Response:
    data = {
        "refresh_token": refresh_token,
        "grant_type": "refresh_token",
        "client_id": client_pair.client_id,
        "client_secret": client_pair.client_secret,
    }

    return requests.post(config.token_url, data=data)


def get_oauth2_config_from(template, sources: ConfiguredOAuth2Sources) -> OAuth2Configuration:
    template_type = template.configuration.type
    if template_type not in sources:
        raise ObjectNotFound(f"oauth information not available for template type {template_type}")
    return sources[template_type]


def read_oauth2_info_from_configuration(
    template_configuration: TemplateConfiguration,
    user_details: UserDetailsDict,
    environment: EnvironmentDict,
) -> Tuple[OAuth2ClientPair, Optional[str]]:
    template_variables = {
        "user": user_details,
        "environment": environment,
    }

    expanded_config = _expand_raw_config(template_configuration, template_variables)
    oauth2_client_id = expanded_config["oauth2_client_id"]
    oauth2_client_secret = expanded_config["oauth2_client_secret"]
    oauth2_scope = cast(Optional[str], expanded_config.get("oauth2_scope"))
    client_pair = OAuth2ClientPair(client_id=oauth2_client_id, client_secret=oauth2_client_secret)
    return client_pair, oauth2_scope


# Things added to configuration dictionary not managed by the template
# but injected dynamically. Currently only `oauth2_access_token`.
class ImplicitConfigurationParameters(TypedDict):
    oauth2_access_token: NotRequired[str]


M = TypeVar("M", bound="BaseModel")


# Implementation copied from https://github.com/pydantic/pydantic/issues/12329#issuecomment-3382159312
def _make_field_optional(field_info: FieldInfo):
    """Returns the field's definition to be used in a `create_model()` call to make the field optional."""
    annotation = field_info.annotation
    assert annotation is not None
    if field_info.is_required():
        return Annotated[Union[annotation, None], field_info], None
    else:
        return Annotated[annotation, field_info]


def make_model_with_all_fields_optional(model: Type[M], fields=None) -> Type[M]:
    """Returns a new Pydantic model based on `model`, but with all fields optional."""
    if fields is None:
        fields = model.model_fields.items()
    return create_model(
        model.__name__,
        __doc__=model.__doc__,
        __base__=model,
        **{field_name: _make_field_optional(field_info) for field_name, field_info in fields},
    )


# TODO: This is a workaround to make all fields optional.
#       It should be removed when Python/pydantic supports this feature natively.
# https://github.com/pydantic/pydantic/issues/1673
def partial_model(
    include: Optional[List[str]] = None, exclude: Optional[List[str]] = None
) -> Callable[[Type[M]], Type[M]]:
    """Decorator to make all model fields optional"""

    if exclude is None:
        exclude = []

    def decorator(model: Type[M]) -> Type[M]:
        if include is None:
            fields: Iterable[tuple[str, FieldInfo]] = model.model_fields.items()
        else:
            fields = ((k, v) for k, v in model.model_fields.items() if k in include)

        if exclude is not None:
            fields = ((k, v) for k, v in fields if k not in exclude)

        return make_model_with_all_fields_optional(model, fields)

    return decorator



"""There is some shared logic between matching/multiplying inputs in workflows
and tools. This module is meant to capture some general permutation logic that
can be applicable for both cases but will only be used in the newer tools case
first.

Maybe this doesn't make sense and maybe much of this stuff could be replaced
with itertools product and permutations. These are open questions.
"""

import copy
from typing import Tuple

from galaxy.exceptions import MessageException
from galaxy.util.bunch import Bunch

input_classification = Bunch(
    SINGLE="single",
    MATCHED="matched",
    MULTIPLIED="multiplied",
)


class InputMatchedException(MessageException):
    """Indicates problem matching inputs while building up inputs
    permutations."""


def build_combos(single_inputs, matched_multi_inputs, multiplied_multi_inputs, nested):
    # Build up every combination of inputs to be run together.
    input_combos = __extend_with_matched_combos(single_inputs, matched_multi_inputs, nested)
    input_combos = __extend_with_multiplied_combos(input_combos, multiplied_multi_inputs, nested)
    return input_combos


def __extend_with_matched_combos(single_inputs, multi_inputs, nested):
    """

    {a => 1, b => 2} and {c => {3, 4}, d => {5, 6}}

    Becomes

    [ {a => 1, b => 2, c => 3, d => 5}, {a => 1, b => 2, c => 4, d => 6}, ]

    """

    if len(multi_inputs) == 0:
        return [single_inputs]

    matched_multi_inputs = []

    first_multi_input_key = next(iter(multi_inputs.keys()))
    first_multi_value = multi_inputs.get(first_multi_input_key)

    for value in first_multi_value:
        new_inputs = __copy_and_extend_inputs(single_inputs, first_multi_input_key, value, nested=nested)
        matched_multi_inputs.append(new_inputs)

    for multi_input_key, multi_input_values in multi_inputs.items():
        if multi_input_key == first_multi_input_key:
            continue
        if len(multi_input_values) != len(first_multi_value):
            raise InputMatchedException(
                f"Received {len(multi_input_values)} inputs for '{multi_input_key}' and {len(first_multi_value)} inputs for '{first_multi_input_key}', these should be of equal length"
            )

        for index, value in enumerate(multi_input_values):
            state_set_value(matched_multi_inputs[index], multi_input_key, value, nested)

    return matched_multi_inputs


def __extend_with_multiplied_combos(input_combos, multi_inputs, nested):
    combos = input_combos

    for multi_input_key, multi_input_value in multi_inputs.items():
        iter_combos = []

        for combo in combos:
            for input_value in multi_input_value:
                iter_combo = __copy_and_extend_inputs(combo, multi_input_key, input_value, nested)
                iter_combos.append(iter_combo)

        combos = iter_combos

    return combos


def __copy_and_extend_inputs(inputs, key, value, nested):
    # can't deepcopy dicts with our models for reason I don't understand,
    # test_map_over_two_collections_unlinked breaks if I try to combine these two branches of the if
    new_inputs = state_copy(inputs, nested)
    state_set_value(new_inputs, key, value, nested)
    return new_inputs


def state_copy(inputs, nested):
    # can't deepcopy dicts with our models for reason I don't understand,
    # test_map_over_two_collections_unlinked breaks if I try to combine these two branches of the if
    if nested:
        state_dict_copy = copy.deepcopy(inputs)
    else:
        state_dict_copy = dict(inputs)
    return state_dict_copy


def state_set_value(state_dict, key, value, nested):
    if "|" not in key or not nested:
        state_dict[key] = value
    else:
        first, rest = key.split("|", 1)
        if first not in state_dict and looks_like_flattened_repeat_key(first):
            repeat_name, index = split_flattened_repeat_key(first)
            if repeat_name not in state_dict:
                state_dict[repeat_name] = []
            repeat_state = state_dict[repeat_name]
            while len(repeat_state) <= index:
                repeat_state.append({})
            state_set_value(repeat_state[index], rest, value, nested)
        else:
            state_set_value(state_dict[first], rest, value, nested)


def state_remove_value(state_dict, key, nested):
    if "|" not in key or not nested:
        del state_dict[key]
    else:
        first, rest = key.split("|", 1)
        child_dict = state_dict[first]
        # repeats?
        if "|" in rest:
            state_remove_value(child_dict, rest, nested)
        else:
            del child_dict[rest]
            if len(child_dict) == 0:
                del state_dict[first]


def state_get_value(state_dict, key, nested):
    if "|" not in key or not nested:
        return state_dict[key]
    else:
        first, rest = key.split("|", 1)
        if first not in state_dict and looks_like_flattened_repeat_key(first):
            repeat_name, index = split_flattened_repeat_key(first)
            return state_get_value(state_dict[repeat_name][index], rest, nested)
        else:
            return state_get_value(state_dict[first], rest, nested)


def is_in_state(state_dict, key, nested):
    if not state_dict:
        return False
    if "|" not in key or not nested:
        return key in state_dict
    else:
        first, rest = key.split("|", 1)
        # repeats?
        is_in_state(state_dict.get(first), rest, nested)


def looks_like_flattened_repeat_key(key: str) -> bool:
    parts = key.rsplit("_", 1)
    return len(parts) == 2 and parts[1].isdigit()


def split_flattened_repeat_key(key: str) -> Tuple[str, int]:
    input_name, _index = key.rsplit("_", 1)
    index = int(_index)
    return input_name, index



import os
from copy import deepcopy
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from galaxy.util import (
    parse_xml,
    unicodify,
)

if TYPE_CHECKING:
    from galaxy.util import (
        Element,
        ElementTree,
    )
    from galaxy.util.path import StrPath

MacrosDictT = Dict[str, List["Element"]]


def load_with_references(path: "StrPath") -> Tuple["ElementTree", Optional[List[str]]]:
    """Load XML documentation from file system and preprocesses XML macros.

    Return the XML representation of the expanded tree and paths to
    referenced files that were imported (macros).
    """
    tree = raw_xml_tree(path)
    root = tree.getroot()

    macros_el = _macros_el(root)
    if macros_el is None:
        return tree, []

    macros: MacrosDictT = {}
    macro_paths = _import_macros(macros_el, path, macros)
    macros_el.clear()

    # Collect tokens
    tokens: Dict[str, str] = {}
    for m in macros.get("token", []):
        token_name = m.get("name")
        assert token_name
        tokens[token_name] = m.text or ""
    tokens = expand_nested_tokens(tokens)

    # Expand xml macros
    macro_dict: Dict[str, XmlMacroDef] = {}
    for m in macros.get("xml", []):
        macro_name = m.get("name")
        assert macro_name
        macro_dict[macro_name] = XmlMacroDef(m)
    _expand_macros([root], macro_dict, tokens)

    # reinsert template macro which are used during tool execution
    for m in macros.get("template", []):
        macros_el.append(m)
    _expand_tokens_for_el(root, tokens)
    return tree, macro_paths


def load(path: "StrPath") -> "ElementTree":
    tree, _ = load_with_references(path)
    return tree


def template_macro_params(root: "Element") -> Dict[str, Union[str, None]]:
    """
    Look for template macros and populate param_dict (for cheetah)
    with these.
    """
    macros_el = _macros_el(root)
    if macros_el is not None:
        return _macros_of_type(macros_el, "template", lambda el: el.text)
    return {}


def raw_xml_tree(path: "StrPath") -> "ElementTree":
    """Load raw (no macro expansion) tree representation of XML represented
    at the specified path.
    """
    tree = parse_xml(path, strip_whitespace=False, remove_comments=True)
    return tree


def imported_macro_paths(root: "Element") -> List[str]:
    macros_el = _macros_el(root)
    if macros_el is None:
        return []
    return _imported_macro_paths_from_el(macros_el)


def _import_macros(macros_el: "Element", path: "StrPath", macros: MacrosDictT) -> Optional[List[str]]:
    """
    root the parsed XML tree
    path the path to the main xml document
    """
    xml_base_dir = os.path.dirname(path)
    macro_paths = _load_macros(macros_el, xml_base_dir, macros)
    # _xml_set_children(macros_el, macro_els)
    return macro_paths


def _macros_el(root: "Element") -> Union["Element", None]:
    return root.find("macros")


T = TypeVar("T")


def _macros_of_type(macros_el: "Element", type: str, el_func: Callable[["Element"], T]) -> Dict[str, T]:
    macro_els = macros_el.findall("macro")
    ret: Dict[str, T] = {}
    for macro_el in macro_els:
        if macro_el.get("type") == type:
            macro_name = macro_el.get("name")
            assert macro_name
            ret[macro_name] = el_func(macro_el)
    return ret


def expand_nested_tokens(tokens: Dict[str, str]) -> Dict[str, str]:
    for token_name in tokens.keys():
        for current_token_name, current_token_value in tokens.items():
            if token_name in current_token_value:
                if token_name == current_token_name:
                    raise Exception(f"Token '{token_name}' cannot contain itself")
                tokens[current_token_name] = current_token_value.replace(token_name, tokens[token_name])
    return tokens


def _expand_tokens(elements: Iterable["Element"], tokens: Dict[str, str]) -> None:
    if not tokens:
        return

    for element in elements:
        _expand_tokens_for_el(element, tokens)


def _expand_tokens_for_el(element: "Element", tokens: Dict[str, str]) -> None:
    """
    expand tokens in element and (recursively) in its children
    replacements of text attributes and attribute values are
    possible
    """
    element_text = element.text
    if element_text:
        new_value = _expand_tokens_str(element_text, tokens)
        if new_value is not element_text:
            element.text = new_value
    for key, value in element.attrib.items():
        new_value = _expand_tokens_str(unicodify(value), tokens)
        if new_value is not value:
            element.attrib[key] = new_value
        new_key = _expand_tokens_str(unicodify(key), tokens)
        if new_key is not key:
            element.attrib[new_key] = element.attrib[key]
            del element.attrib[key]
    # recursively expand in childrens
    _expand_tokens(element.__iter__(), tokens)


def _expand_tokens_str(s: str, tokens: Dict[str, str]) -> str:
    for key, value in tokens.items():
        if key in s:
            s = s.replace(key, value)
    return s


def _expand_macros(
    elements: Iterable["Element"],
    macros: Dict[str, "XmlMacroDef"],
    tokens: Dict[str, str],
    visited: Optional[List[str]] = None,
) -> None:
    if not macros and not tokens:
        return

    if visited is None:
        visited = []

    for element in elements:
        while True:
            expand_el = element.find(".//expand")
            if expand_el is None:
                break
            _expand_macro(expand_el, macros, tokens, visited)


def _expand_macro(
    expand_el: "Element", macros: Dict[str, "XmlMacroDef"], tokens: Dict[str, str], visited: List[str]
) -> None:
    macro_name = expand_el.get("macro")
    assert macro_name is not None, "Attempted to expand macro with no 'macro' attribute defined."
    # check for cycles in the nested macro expansion
    assert (
        macro_name not in visited
    ), f"Cycle in nested macros: already expanded {visited} can't expand '{macro_name}' again"
    visited.append(macro_name)

    assert macro_name in macros, f"No macro named {macro_name} found, known macros are {', '.join(macros.keys())}."
    macro_def = macros[macro_name]
    macro_el = deepcopy(macro_def.element)
    _expand_yield_statements(macro_el, expand_el)

    macro_tokens = macro_def.macro_tokens(expand_el)
    if macro_tokens:
        _expand_tokens(macro_el.__iter__(), macro_tokens)

    # Recursively expand contained macros.
    _expand_macros(macro_el.__iter__(), macros, tokens, visited)
    _xml_replace(expand_el, macro_el.__iter__())
    del visited[-1]


def _expand_yield_statements(macro_el: "Element", expand_el: "Element") -> None:
    """
    Modifies the macro_el element by replacing
    1. all named yield tags by the content of the corresponding token tags
       - token tags need to be direct children of the expand
       - processed in order of definition of the token tags
    2. all unnamed yield tags by the non-token children of the expand tag
    """
    # replace named yields
    for token_el in expand_el.findall("./token"):
        name = token_el.attrib.get("name")
        assert name is not None, "Found unnamed token" + str(token_el.attrib)
        yield_els = list(macro_el.findall(f".//yield[@name='{name}']"))
        assert len(yield_els) > 0, f"No named yield found for named token {name}"
        for yield_el in yield_els:
            _xml_replace(yield_el, token_el.__iter__())

    # replace unnamed yields
    yield_els = list(macro_el.findall(".//yield"))
    expand_el_children = [c for c in expand_el if c.tag != "token"]
    for yield_el in yield_els:
        _xml_replace(yield_el, expand_el_children)


def _load_macros(macros_el: "Element", xml_base_dir: str, macros: MacrosDictT) -> List[str]:
    # Import macros from external files.
    macro_paths = _load_imported_macros(macros_el, xml_base_dir, macros)
    # Load all directly defined macros.
    _load_embedded_macros(macros_el, macros)
    return macro_paths


def _load_embedded_macros(macros_el: "Element", macros: MacrosDictT) -> None:
    # attribute typed macro
    for macro in macros_el.iterfind("macro"):
        if "type" not in macro.attrib:
            macro.attrib["type"] = "xml"
        macro_type = unicodify(macro.attrib["type"])
        try:
            macros[macro_type].append(macro)
        except KeyError:
            macros[macro_type] = [macro]

    # type shortcuts (<xml> is a shortcut for <macro type="xml",
    # likewise for <template>.
    for tag in ["template", "xml", "token"]:
        for macro_el in macros_el.iterfind(tag):
            macro_el.attrib["type"] = tag
            macro_el.tag = "macro"
            try:
                macros[tag].append(macro_el)
            except KeyError:
                macros[tag] = [macro_el]


def _load_imported_macros(macros_el: "Element", xml_base_dir: str, macros: MacrosDictT) -> List[str]:
    macro_paths = []

    for tool_relative_import_path in _imported_macro_paths_from_el(macros_el):
        import_path = os.path.join(xml_base_dir, tool_relative_import_path)
        macro_paths.append(import_path)
        current_macro_paths = _load_macro_file(import_path, xml_base_dir, macros)
        macro_paths.extend(current_macro_paths)
    return macro_paths


def _imported_macro_paths_from_el(macros_el: "Element") -> List[str]:
    imported_macro_paths = []
    for macro_import_el in macros_el.findall("import"):
        raw_import_path = macro_import_el.text
        assert raw_import_path
        imported_macro_paths.append(raw_import_path)
    return imported_macro_paths


def _load_macro_file(path: "StrPath", xml_base_dir: str, macros: MacrosDictT) -> List[str]:
    tree = parse_xml(path, strip_whitespace=False)
    root = tree.getroot()
    return _load_macros(root, xml_base_dir, macros)


def _xml_replace(query: "Element", targets: Iterable["Element"]) -> None:
    parent_el = query.find("..")
    assert parent_el is not None
    matching_index = -1
    # for index, el in enumerate(parent_el.iter('.')):  ## Something like this for newer implementation
    for index, el in enumerate(parent_el):
        if el == query:
            matching_index = index
            break
    assert matching_index >= 0
    current_index = matching_index
    for target in targets:
        current_index += 1
        parent_el.insert(current_index, deepcopy(target))
    parent_el.remove(query)


class XmlMacroDef:
    """
    representation of a (Galaxy) XML macro

    stores the root element of the macro and the parameters.
    each parameter is represented as pair containing
    - the quote character, default '@'
    - parameter name

    Parameter names can be given as comma separated list using the
    `tokens` attribute or as attributes `token_XXX` (where `XXX` is the name).
    The former option should be used to specify required attributes of the
    macro and the latter for optional attributes of the macro (the value of
    `token_XXX is used as default value).

    TODO: `token_quote` forbids `"quote"` as character name of optional
    parameters
    """

    def __init__(self, el: "Element") -> None:
        self.element = el
        tokens: Dict[str, Union[str, None]] = {}
        self.token_quote = "@"
        for key, value in el.attrib.items():
            key = unicodify(key)
            value = unicodify(value)
            if key == "token_quote":
                self.token_quote = value
            if key == "tokens":
                for token in value.split(","):
                    tokens[token] = None  # here None means that the token is a required parameter
            elif key.startswith("token_"):
                token = key[len("token_") :]
                tokens[token] = value
        self.tokens = tokens

    def macro_tokens(self, expand_el: "Element") -> Dict[str, str]:
        """
        get a dictionary mapping token names to values. The names are the
        parameter names surrounded by the quote character. Values are taken
        from the expand_el if absent default values of optional parameters are
        used.
        """
        tokens: Dict[str, str] = {}
        for key, default_val in self.tokens.items():
            token_value = expand_el.attrib.get(key, default_val)
            if token_value is None:
                raise ValueError(f"Failed to expand macro - missing required parameter [{key}].")
            token_name = f"{self.token_quote}{key.upper()}{self.token_quote}"
            tokens[token_name] = token_value
        return tokens


__all__ = (
    "imported_macro_paths",
    "load",
    "load_with_references",
    "raw_xml_tree",
    "template_macro_params",
)



"""Entry point for the usage of Cheetah templating within Galaxy."""

import sys
import traceback
from typing import (
    Optional,
    Union,
)

from Cheetah.Compiler import Compiler
from Cheetah.NameMapper import NotFound
from Cheetah.Parser import ParseError
from Cheetah.Template import Template
from packaging.version import Version

from galaxy.util.tree_dict import TreeDict
from . import unicodify

try:
    from lib2to3.refactor import RefactoringTool
except ImportError:
    # Either Python 3.13 or Debian(<=12)/Ubuntu(<=24.10) without the
    # python3-lib2to3 package
    import fissix
    from fissix import (
        fixes as fissix_fixes,
        pgen2 as fissix_pgen2,
        refactor as fissix_refactor,
    )

    sys.modules["lib2to3"] = fissix
    sys.modules["lib2to3.fixes"] = fissix_fixes
    sys.modules["lib2to3.pgen2"] = fissix_pgen2
    sys.modules["lib2to3.refactor"] = fissix_refactor

    from lib2to3.refactor import RefactoringTool

from past.translation import myfixes

# Skip libpasteurize fixers, which make sure code is py2 and py3 compatible.
# This is not needed, we only translate code on py3.
myfixes = [f for f in myfixes if not f.startswith("libpasteurize")]
refactoring_tool = RefactoringTool(myfixes, {"print_function": True})


class InputNotFoundSyntaxError(SyntaxError):
    pass


class FixedModuleCodeCompiler(Compiler):
    module_code = None

    def getModuleCode(self):
        self._moduleDef = self.module_code
        return self._moduleDef


def create_compiler_class(module_code):
    class CustomCompilerClass(FixedModuleCodeCompiler):
        pass

    CustomCompilerClass.module_code = module_code
    return CustomCompilerClass


def fill_template(
    template_text,
    context=None,
    retry=10,
    compiler_class=Compiler,
    first_exception=None,
    futurized=False,
    python_template_version: Optional[Union[str, Version]] = "3",
    **kwargs,
):
    """Fill a cheetah template out for specified context.

    If template_text is None, an exception will be thrown, if context
    is None (the default) - keyword arguments to this function will be used
    as the context.
    """
    if template_text is None:
        raise TypeError("Template text specified as None to fill_template.")
    if not context:
        context = kwargs
    if python_template_version is None:
        python_template_version = Version("3")
    elif isinstance(python_template_version, str):
        python_template_version = Version(python_template_version)
    try:
        klass = Template.compile(source=template_text, compilerClass=compiler_class)
    except ParseError as e:
        # Might happen on invalid syntax within a cheetah statement, like `#if $smxsize <> 128.0`
        if first_exception is None:
            first_exception = e
        if python_template_version.release[0] < 3 and retry > 0:
            module_code = Template.compile(
                source=template_text, compilerClass=compiler_class, returnAClass=False
            ).decode("utf-8")
            module_code = futurize_preprocessor(module_code)
            compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry - 1,
                compiler_class=compiler_class,
                first_exception=first_exception,
                python_template_version=python_template_version,
            )
        raise first_exception or e
    t = klass(searchList=[context])
    try:
        return unicodify(t, log_exception=False)
    except (NotFound, InputNotFoundSyntaxError) as e:
        if first_exception is None:
            first_exception = e
        if not isinstance(context, TreeDict):
            masked_input = None
            if "input" in context and callable(context["input"]):
                masked_input = context.pop("input", None)
            context = TreeDict(context)
            if "input" not in context and masked_input:
                context["input"] = masked_input
        tb = e.__traceback__
        if retry > 0:
            if python_template_version.release[0] < 3:
                last_stack = traceback.extract_tb(tb)[-1]
                if last_stack.name == "<listcomp>" and last_stack.lineno:
                    # On python 3 list, dict and set comprehensions as well as generator expressions
                    # have their own local scope, which prevents accessing frame variables in cheetah.
                    # We can work around this by replacing `$var` with `var`, but we only do this for
                    # list comprehensions, as this has never worked for dict or set comprehensions or
                    # generator expressions in Cheetah.
                    var_not_found = e.args[0].split("'")[1]
                    replace_str = f'VFFSL(SL,"{var_not_found}",True)'
                    lineno = last_stack.lineno - 1
                    module_code = t._CHEETAH_generatedModuleCode.splitlines()
                    module_code[lineno] = module_code[lineno].replace(replace_str, var_not_found)
                    module_code = "\n".join(module_code)
                    compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry - 1,
                compiler_class=compiler_class,
                first_exception=first_exception,
                python_template_version=python_template_version,
            )
        raise first_exception or e
    except Exception as e:
        if first_exception is None:
            first_exception = e
        if python_template_version.release[0] < 3 and not futurized:
            # Possibly an error caused by attempting to run python 2
            # template code on python 3. Run the generated module code
            # through futurize and hope for the best.
            module_code = t._CHEETAH_generatedModuleCode
            module_code = futurize_preprocessor(module_code)
            compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry,
                compiler_class=compiler_class,
                first_exception=first_exception,
                futurized=True,
                python_template_version=python_template_version,
            )
        raise first_exception or e


def futurize_preprocessor(source):
    source = str(refactoring_tool.refactor_string(source, name="auto_translate_cheetah"))
    # libfuturize.fixes.fix_unicode_keep_u' breaks from Cheetah.compat import unicode
    source = source.replace("from Cheetah.compat import str", "from Cheetah.compat import unicode")
    return source



import logging
from typing import (
    Any,
    cast,
    Optional,
)


class GalaxyLogger(logging.Logger):
    def trace(self, message: object, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(LOGLV_TRACE):
            self._log(LOGLV_TRACE, message, args, **kwargs)


# Add custom "TRACE" log level for ludicrous verbosity.
LOGLV_TRACE = 5
logging.addLevelName(LOGLV_TRACE, "TRACE")
logging.setLoggerClass(GalaxyLogger)


def get_logger(name: Optional[str] = None) -> GalaxyLogger:
    logger = logging.getLogger(name)
    return cast(GalaxyLogger, logger)
